(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2000 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* Based on Trevor Jim's ocamldot.mll *)

{
structure StringSet = MakeSet(struct type t = string
                                  val compare = compare
                           end)

(* implode : char list -> string *)
fun implode chars =
  let val buf = String.create (List.length chars)
      fun implode' (i,l) =
        case l
          of [] => buf
           | x::tl => (String.set buf i x; implode' (i+1,tl))
  in  implode' (0, chars)
  end

(* explode : string -> char list *)
fun explode str =
  let fun f (chars, -1) = chars
        | f (chars, i) = f ((String.get str i)::chars, i-1)
  in  f (nil, String.length str - 1)
  end

fun isAlpha c =
  let val i = Char.code c
  in  (i >= 65 andalso i <= 90) orelse (i >= 97 andalso i <= 122)
  end

fun dropdots string =
  let fun drop nil = nil
        | drop (c::rest) = if isAlpha c then (c::rest) else drop rest
  in  implode(drop (explode string))
  end

val dependencies = ref []
val currentSource = ref ""
fun addDepend t = let val s = !currentSource
                  in  if s<>t
                      then dependencies := (s,t)::(!dependencies)
                  end
}

rule processSource = parse
    ['.' '-' '/' 'A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246'
     '\248'-'\255' '\'' '0'-'9' ]+ '.' ['A'-'Z' 'a'-'z']+
    [' ' '\009']* ':'
      { let val s = Lexing.lexeme lexbuf
            val s = dropdots s
            val i = String.index s '.'
            val s = String.sub s 0 i
            val s = Filename.basename s
            val s = String.capitalize s
        in  currentSource := s; processTargets lexbuf
        end }
  | eof
      { () }
  | _
      { processSource lexbuf }

and processTargets = parse
    [' ' '\009']+
      { processTargets lexbuf }
  | '\\' [' ' '\009']* ['\010' '\013']+ [' ' '\009']+
      { processTargets lexbuf }
  | ['.' '/' 'A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246'
     '\248'-'\255' '\'' '0'-'9' ]+ '.' ['A'-'Z' 'a'-'z']+
      { let val t = Lexing.lexeme lexbuf
            val t = dropdots t
            val i = String.index t '.'
            val t = String.sub t 0 i
            val t = Filename.basename t
            val t = String.capitalize t
        in  addDepend t; processTargets lexbuf
        end }
  | eof
      { () }
  | _
      { processSource lexbuf }

{
                  (********************************)
                  (* Utility functions for graphs *)
                  (********************************)

(**********************************************************************)
(* A graph is represented by a (string * StringSet) list,             *)
(* that is, a list of (source,targets) pairs.                         *)
(**********************************************************************)
val emptyGraph = []

(**********************************************************************)
(* divideGraph graph source = (sourceTargets, graphWithoutSource)     *)
(*                                                                    *)
(* Return the targets of a source in a graph and the graph with the   *)
(* source substracted from the sources.  GraphWithoutSources may      *)
(* still contain source as a target.                                  *)
(**********************************************************************)
fun divideGraph graph source =
  let fun aux [] = (StringSet.empty,[])
        | aux ((s,ts)::tl) =
          if s=source then (ts,tl)
          else let val (sourceTargets,tlWithoutSource) = aux tl
               in  (sourceTargets,(s,ts)::tlWithoutSource)
               end
  in  aux graph
  end

(*********************************************)
(* Add the edge (source,target) to the graph *)
(*********************************************)
fun addEdge graph source target =
  let val (sourceTargets,graphWithoutSource) = divideGraph graph source
  in  (source,StringSet.add target sourceTargets)::graphWithoutSource
  end

(************************************************************)
(* Add the edges { (source,t) | t in targets } to the graph *)
(************************************************************)
fun addEdges graph source targets =
  let val (sourceTargets,graphWithoutSource) = divideGraph graph source
  in  (source,StringSet.union targets sourceTargets)::graphWithoutSource
  end

(**************************************************)
(* Remove the edge (source,target) from the graph *)
(**************************************************)
fun removeEdge graph source target =
  let fun loop [] = []
        | loop ((s,ts)::tl) =
          if s=source
          then (s,StringSet.remove target ts)::tl else (s,ts)::(loop tl)
  in  loop graph
  end

(*****************************************************************)
(* Remove the edges { (source,t) | t in targets } from the graph *)
(*****************************************************************)
fun removeEdges graph source targets =
  let fun loop [] = []
        | loop ((s,ts)::tl) =
          if s=source
          then (s,StringSet.diff ts targets)::tl else (s,ts)::(loop tl)
  in  loop graph
  end

(**********************************************************************)
(* Convert between an edge-list representation of graphs and our      *)
(* representation.                                                    *)
(**********************************************************************)
fun edgesOfGraph graph =
  List.concat
    (List.map
       (fn (s,ts) => List.map (fn t => (s,t)) (StringSet.elements ts)) graph)

fun graphOfEdges edges =
  List.fold_left (fn g => fn (s,t) => addEdge g s t) emptyGraph edges

(****************************)
(* Is an edge in the graph? *)
(****************************)
fun isEdge graph source target =
  let val sourceTargets = List.assoc source graph
      handle Not_found => StringSet.empty
  in  StringSet.mem target sourceTargets
  end handle Not_found => false

(*****************)
(* Print a graph *)
(*****************)
fun printGraph graph =
  let fun printEdges (source,targets) =
        StringSet.iter
          (fn t => Printf.printf "  \"%s\" -> \"%s\" ;\n" source t) targets
  in  List.iter printEdges graph
  end

(********************************)
(* Targets of a node in a graph *)
(********************************)
fun targetsOf graph node = (* A set of nodes *)
  (List.assoc node graph) handle Not_found => StringSet.empty

(*****************************************)
(* Sources that target a node in a graph *)
(*****************************************)
fun sourcesOf graph node = (* A list of nodes *)
  let fun aux [] = []
        | aux ((s,ts)::tl) =
          if StringSet.mem node ts then s::(aux tl) else aux tl
  in  aux graph
  end

(******************************************************************)
(* Add an edge to a transitively closed graph, and return the new *)
(* transitive closure.                                            *)
(******************************************************************)
fun addEdgeTc graph source target =
  let val targetTargets = targetsOf graph target
      val (sourceTargets,graphWithoutSource) = divideGraph graph source
      val sourceSources = sourcesOf graphWithoutSource source
      val newSourceTargets = StringSet.add target
                               (StringSet.union sourceTargets targetTargets)
  in  (source,newSourceTargets)::
      (List.fold_right (fn s g => addEdges g s newSourceTargets)
                       sourceSources graphWithoutSource)
  end

(**********************************************************)
(* Compute the transitive closure of a graph from scratch *)
(**********************************************************)
fun tc graph =
  let fun loop graph (source,targets) =
        let val reachableFromSource =
                List.fold_left (fn r (s,ts) =>
                                if StringSet.mem s r then StringSet.union r ts
                                else r) targets graph
        in  (source,reachableFromSource)::
            (List.map (fn (s,ts) =>
                          if StringSet.mem source ts
                          then (s,StringSet.union ts reachableFromSource)
                          else (s,ts)) graph)
        end
  in  List.fold_left loop [] graph
  end

(************************************************************************)
(* The transitive kernel (tk) of a dag is a subset of the dag whose     *)
(* transitive closure is the same as the transitive closure of the dag. *)
(*                                                                      *)
(* IF THE GRAPH IS NOT A DAG, THIS CODE WON'T WORK PROPERLY!!!          *)
(************************************************************************)

(************************************************************************)
(* Add an edge to a kernel dag and return the new kernel and transitive *)
(* closure of the new kernel.  Requires the transitive closure of the   *)
(* old kernel.                                                          *)
(************************************************************************)
fun addEdgeTk kernel tcKernel source target =
  if isEdge tcKernel source target
  then (kernel,tcKernel)
  else if source=target
       then (addEdge kernel source target,tcKernel)
       else let val (sourceTargets,kernelWithoutSource) =
                    divideGraph kernel source
                val targetTargets =
                    StringSet.add target (targetsOf tcKernel target)
                val sourceSources = sourcesOf tcKernel source
                val kernelWithoutSource =
                    List.fold_left (fn kws s =>
                                       removeEdges kws s targetTargets)
                                   kernelWithoutSource sourceSources
            in  ((source,
                  StringSet.add target
                               (StringSet.diff sourceTargets targetTargets))::
                 kernelWithoutSource, addEdgeTc tcKernel source target)
            end

(**********************************)
(* The transitive kernel of a dag *)
(**********************************)
fun tk dag =
  let val edges = edgesOfGraph dag
      val (kernel,tcKernel) =
          List.fold_left (fn (k,tck) (s,t) => addEdgeTk k tck s t)
                         (emptyGraph,emptyGraph) edges
  in  kernel
  end

(**************************)
(* Print the dependencies *)
(**************************)

val doKernel = ref true
fun printDepend graph =
  if (!doKernel) then printGraph (tk graph) else printGraph graph

val calledOnFile = ref false
fun getDependFromFile file =
  let val _ = calledOnFile := true
      val ic = open_in file
      val lexbuf = Lexing.from_channel ic
  in  processSource lexbuf; close_in ic
  end handle Sys_error msg => ()
           | Exit => ()

fun getDependFromStdin () =
  let val lexbuf = Lexing.from_channel stdin
  in  processSource lexbuf
  end handle Sys_error msg => ()
           | Exit => ()

                          (***************)
                          (* Entry point *)
                          (***************)

val usage = "Usage: ocamldot [options] <files>"

val leftToRight = ref false
val landscape = ref false
val roots = ref []

val _ = Arg.parse
  [ ("-fullgraph",
     Arg.Clear doKernel,
     "  draw the full graph (default is to draw only the kernel)"),
    ("-landscape",
     Arg.Set landscape,
     "  output in landscape format (default is portrait)"),
    ("-lr",
     Arg.Set leftToRight,
     "         draw graph from left to right (default is top to bottom)"),
    ("-r",
     Arg.String (fn s => roots := s::!roots),
     "<r>       use <r> as a root in the graph; nodes reachable from <r>\n               will be shown")
  ] getDependFromFile usage

val _ = (if not(!calledOnFile) then getDependFromStdin();
         print_string "digraph G {\n";
         if !landscape
         then print_string "  size=\"10,7.5\" ;\n  rotate=90 ;\n"
         else print_string "  size=\"7.5,10\" ;\n";
         if (!leftToRight) then print_string "  rankdir = LR ;\n"
         else print_string "  rankdir = TB ;\n")

val graph = graphOfEdges(!dependencies)

val _ = case !roots
          of [] => printDepend graph
           | roots =>
    (* Set up the graph so that the roots are printed at the same level *)
             (print_string "  { rank=same ;\n";
              List.iter (fn r => (print_string ("    "^r^" ;\n"))) roots;
              print_string "  };\n")

(* Find the graph reachable from the roots *)
val tcGraph = tc graph
fun reachable node =
  (List.exists (fn r => r=node) (!roots))
  orelse
  (List.exists (fn r => isEdge tcGraph r node) (!roots))

val reachableFromRoots =
    List.concat
      (List.map (fn (source,targets) => if reachable source
                                        then [(source,targets)] else [])
                 graph)

val _ = (printDepend reachableFromRoots;
         print_string "}\n")

}
